from .multithreaded_offset_generator import MultithreadedOffsetGenerator
from ..helpers.datetime_utils import datetime_to_utc_str, str_to_localized_datetime, utc_now


class InstallmentsGenerator(MultithreadedOffsetGenerator):
    def __init__(self, stream_name, client, config, state, sub_type):
        super(InstallmentsGenerator, self).__init__(stream_name, client, config, state, sub_type)
        self.date_windowing = False

    def _init_endpoint_config(self):
        super(InstallmentsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "installments"
        self.endpoint_api_method = "GET"
        self.endpoint_params = {
            "dueFrom": datetime_to_utc_str(str_to_localized_datetime(self.start_date))[:10],
            "dueTo": datetime_to_utc_str(utc_now())[:10],
            "detailsLevel": "FULL",
            "paginationDetails": "OFF"
        }
        self.endpoint_bookmark_field = "lastPaidDate"

    def transform_batch(self, batch):
        temp_batch = super(InstallmentsGenerator, self).transform_batch(batch)
        for record in temp_batch:
            if "number" in record:
                record["number"] = "0"
        return temp_batch



from abc import ABC
from typing import List
from singer import utils, get_logger

from ..helpers import transform_json
from ..helpers.hashable_dict import HashableDict

LOGGER = get_logger()


class TapGenerator(ABC):
    def __init__(self, stream_name, client, config, state, sub_type):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        self.date_windowing = False
        self.date_window_size = client.window_size
        self.start_windows_datetime_str = None
        self.sub_stream_name = stream_name

        # Define parameters inside init
        self.params = dict()
        self.time_extracted = 0
        self.offset = 0

        # Initialize parameters
        self._init_config()
        self._init_endpoint_config()
        self._init_endpoint_body()
        self._init_buffers()
        self._init_params()

    def _init_config(self):
        self.start_date = self.config.get('start_date')

    def _init_endpoint_config(self):
        self.endpoint_path = ""
        self.endpoint_api_version = "v2"
        self.endpoint_api_method = "POST"
        self.endpoint_params = {
            "detailsLevel": "FULL",
            "paginationDetails": "OFF"
        }
        self.endpoint_sorting_criteria = {
            "field": "encodedKey",
            "order": "ASC"
        }
        self.endpoint_filter_criteria = []
        self.endpoint_api_key_type = None
        self.endpoint_bookmark_field = ""

    def _init_endpoint_body(self):
        self.endpoint_body = {"sortingCriteria": self.endpoint_sorting_criteria,
                              "filterCriteria": self.endpoint_filter_criteria}

    def _init_buffers(self):
        self.buffer: List = list()
        self.max_buffer_size = 10000

    def _init_params(self):
        self.time_extracted = None
        self.static_params = dict(self.endpoint_params)
        self.offset = 0
        self.limit = self.client.page_size
        self.params = self.static_params

    def _all_fetch_batch_steps(self):
        # Large buffer size can impact memory utilization of connector
        # so empty the buffer once it reaches default max limit
        if len(self.buffer) > self.max_buffer_size:
            return

        self.prepare_batch()
        raw_batch = self.fetch_batch()
        self.buffer = transform_json(raw_batch, self.stream_name)
        if not self.buffer:
            LOGGER.warning(f'(generator) Stream {self.stream_name} - NO TRANSFORMED DATA RESULTS')
        self.last_batch_size = len(self.buffer)

    def __iter__(self):
        self._all_fetch_batch_steps()
        return self

    def next(self):
        if not self.buffer:
            if self.last_batch_size < self.limit:
                raise StopIteration()
            self.offset += self.limit
            self._all_fetch_batch_steps()
            if not self.buffer:
                raise StopIteration()
        return self.buffer.pop(0)

    def __next__(self):
        return self.next()

    def prepare_batch(self):
        self.params = {
            "offset": self.offset,
            "limit": self.limit,
            **self.static_params
        }

    def transform_batch(self, batch):
        # Check if batch is list or dict and convert to Hashable dict accordingly
        if type(batch) == list:
            return list(map(HashableDict, batch))
        return HashableDict(batch)

    def fetch_batch(self):
        endpoint_querystring = '&'.join([f'{key}={value}' for (key, value) in self.params.items()])

        LOGGER.info(f'(generator) Stream {self.stream_name} - URL for {self.stream_name} ({self.endpoint_api_method}, '
                    f'{self.endpoint_api_version}): {self.client.base_url}/{self.endpoint_path}?{endpoint_querystring}')
        LOGGER.info(f'(generator) Stream {self.stream_name} - body = {self.endpoint_body}')

        response = self.client.request(
            method=self.endpoint_api_method,
            path=self.endpoint_path,
            version=self.endpoint_api_version,
            apikey_type=self.endpoint_api_key_type,
            params=endpoint_querystring,
            endpoint=self.stream_name,
            json=self.endpoint_body
        )

        self.time_extracted = utils.now()
        LOGGER.info(f'(generator) Stream {self.stream_name} - extracted records: {len(response)}')
        return self.transform_batch(response)

    def get_default_start_value(self):
        return None

    def set_last_sync_completed(self, end_time):
        pass

    def wait_for_slibling_to_catchup(self):
        pass



from mock import MagicMock

from tap_mambu.tap_processors.child_processor import ChildProcessor
from ..constants import config_json
from ..helpers import GeneratorMock
from tap_mambu.helpers.generator_processor_pairs import get_generator_processor_for_stream


def setup_processor_base_test(stream_name):
    from tap_mambu import discover
    catalog = discover()
    client_mock = MagicMock()
    _, processor_class = get_generator_processor_for_stream(stream_name)

    processor = processor_class(catalog=catalog,
                                stream_name=stream_name,
                                client=client_mock,
                                config=config_json,
                                state={"currently_syncing": stream_name},
                                sub_type="self",
                                generators=[GeneratorMock([])],
                                **({"parent_id": "TEST"} if issubclass(processor_class, ChildProcessor) else {}))
    return processor



from . import setup_generator_base_test


def test_deposit_transactions_generator():
    generators = setup_generator_base_test("deposit_transactions")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "deposits/transactions:search"
    assert generator.endpoint_bookmark_field == "creationDate"
    assert generator.endpoint_sorting_criteria == {
            "field": "id",
            "order": "ASC"
        }
    assert generator.endpoint_filter_criteria == [
            {
                "field": "creationDate",
                "operator": "AFTER",
                "value": '2021-06-01T00:00:00.000000Z'
            }
        ]



from . import setup_generator_base_test


def test_interest_accrual_breakdown_generator_endpoint_config_init():
    generators = setup_generator_base_test("interest_accrual_breakdown")

    assert 1 == len(generators)
    generator = generators[0]

    assert generator.endpoint_path == 'accounting/interestaccrual:search'
    assert generator.endpoint_bookmark_field == "creationDate"
    assert generator.endpoint_sorting_criteria == {
        "field": "entryId",
        "order": "ASC"
    }
    assert generator.endpoint_filter_criteria == [
        {
            "field": "creationDate",
            "operator": "AFTER",
            "value": '2021-06-01'
        }
    ]



import singer

from .helpers.constants import DEFAULT_PAGE_SIZE
from .helpers import get_selected_streams, should_sync_stream, update_currently_syncing
from .helpers.datetime_utils import get_timezone_info
from .helpers.generator_processor_pairs import get_generator_processor_for_stream, get_stream_subtypes
from .helpers.multithreaded_requests import MultithreadedRequestsPool

LOGGER = singer.get_logger()


def sync_endpoint(client, catalog, state,
                  stream_name, sub_type, config, parent_id=None):
    generator_classes, processor_class = get_generator_processor_for_stream(stream_name)
    generators = [generator_class(stream_name=stream_name,
                                  client=client,
                                  config=config,
                                  state=state,
                                  sub_type=sub_type,
                                  **({} if parent_id is None else {"parent_id": parent_id}))
                  for generator_class in generator_classes]
    processor = processor_class(catalog=catalog,
                                stream_name=stream_name,
                                client=client,
                                config=config,
                                state=state,
                                sub_type=sub_type,
                                generators=generators,
                                **({} if parent_id is None else {"parent_id": parent_id}))

    return processor.process_streams_from_generators()


def sync_all_streams(client, config, catalog, state):
    from .tap_generators.child_generator import ChildGenerator
    from .tap_processors.child_processor import ChildProcessor

    get_timezone_info(client)

    selected_streams = get_selected_streams(catalog)
    LOGGER.info('selected_streams: {}'.format(selected_streams))

    if not selected_streams:
        return

    # last_stream = Previous currently synced stream, if the load was interrupted
    last_stream = singer.get_currently_syncing(state)
    LOGGER.info('last/currently syncing stream: {}'.format(last_stream))

    # Start syncing from last/currently syncing stream
    if last_stream in selected_streams:
        selected_streams = selected_streams[selected_streams.index(last_stream):] +\
                           selected_streams[:selected_streams.index(last_stream)]

    # For each endpoint (above), determine if the stream should be streamed
    #   (based on the catalog and last_stream), then sync those streams.
    for stream_name in selected_streams:
        # Check if stream is child stream (and ignore it)
        generators, processor = get_generator_processor_for_stream(stream_name)
        if any([issubclass(generator, ChildGenerator) for generator in generators]) or \
                issubclass(processor, ChildProcessor):
            continue

        should_stream, last_stream = should_sync_stream(selected_streams,
                                                        last_stream,
                                                        stream_name)

        if should_stream:
            # loop through each sub type
            sub_types = get_stream_subtypes(stream_name)
            for sub_type in sub_types:
                LOGGER.info('START Syncing: {}, Type: {}'.format(stream_name, sub_type))

                update_currently_syncing(state, stream_name)
                total_records = sync_endpoint(
                    client=client,
                    catalog=catalog,
                    state=state,
                    stream_name=stream_name,
                    sub_type=sub_type,
                    config=config
                )

                update_currently_syncing(state, None)
                LOGGER.info('Synced: {}, total_records: {}'.format(
                                stream_name,
                                total_records))
                LOGGER.info('FINISHED Syncing: {}'.format(stream_name))

    MultithreadedRequestsPool.shutdown()



from .multithreaded_bookmark_generator import MultithreadedBookmarkGenerator
from ..helpers.datetime_utils import datetime_to_utc_str


class LoanTransactionsGenerator(MultithreadedBookmarkGenerator):
    def _init_endpoint_config(self):
        super(LoanTransactionsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "loans/transactions:search"
        self.endpoint_bookmark_field = "creationDate"
        self.endpoint_sorting_criteria["field"] = "id"

    def prepare_batch_params(self):
        super(LoanTransactionsGenerator, self).prepare_batch_params()
        self.endpoint_filter_criteria[0]["value"] = datetime_to_utc_str(self.endpoint_intermediary_bookmark_value)



from .multithreaded_bookmark_generator import MultithreadedBookmarkGenerator
from ..helpers.datetime_utils import datetime_to_utc_str, str_to_datetime
from ..helpers import get_bookmark, write_bookmark


class LoanAccountsLMGenerator(MultithreadedBookmarkGenerator):
    def __init__(self, stream_name, client, config, state, sub_type):
        super(LoanAccountsLMGenerator, self).__init__(stream_name, client, config, state, sub_type)
        self.max_threads = 3
        self.sub_stream_name = "loan_accounts_lmg"

    def _init_endpoint_config(self):
        super(LoanAccountsLMGenerator, self)._init_endpoint_config()
        self.endpoint_path = "loans:search"
        self.endpoint_bookmark_field = "lastModifiedDate"
        self.endpoint_sorting_criteria = {
            "field": "id",
            "order": "ASC"
        }

    def prepare_batch_params(self):
        super(LoanAccountsLMGenerator, self).prepare_batch_params()
        self.endpoint_filter_criteria[0]["value"] = datetime_to_utc_str(
            self.endpoint_intermediary_bookmark_value)

    def write_sub_stream_bookmark(self, start):
        write_bookmark(self.state, self.sub_stream_name, self.sub_type, start)

    def get_default_start_value(self):
        # Historical sync will use start date as as date window
        # Increamental syncs will use last stream bookmark value
        # Interrupted syncs will use last winodow of sub-stream as first date window
        stream_bookmark = get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)
        sub_stream_bookmark = get_bookmark(self.state, self.sub_stream_name, self.sub_type, stream_bookmark)
        if self.compare_bookmark_values(sub_stream_bookmark, stream_bookmark):
            start_value = stream_bookmark
        else:
            start_value = sub_stream_bookmark
        truncated_start_date = datetime_to_utc_str(str_to_datetime(start_value).replace(hour=0, minute=0, second=0))
        return truncated_start_date

    def remove_sub_stream_bookmark(self):
        # Remove sub-stream bookmark once we finish extraction till current date
        if self.sub_stream_name in self.state.get("bookmarks", {}):
            del self.state["bookmarks"][self.sub_stream_name]

    def set_intermediary_bookmark(self, record_bookmark_value):
        if self.endpoint_intermediary_bookmark_value is None or \
                self.compare_bookmark_values(record_bookmark_value,
                                             self.endpoint_intermediary_bookmark_value):
            self.endpoint_intermediary_bookmark_offset = 1
            return

        if record_bookmark_value == self.endpoint_intermediary_bookmark_value:
            self.endpoint_intermediary_bookmark_offset += 1
            return


class LoanAccountsADGenerator(LoanAccountsLMGenerator):
    def __init__(self, stream_name, client, config, state, sub_type):
        super(LoanAccountsADGenerator, self).__init__(stream_name, client, config, state, sub_type)
        self.sub_stream_name = "loan_accounts_adg"

    def _init_endpoint_config(self):
        super(LoanAccountsADGenerator, self)._init_endpoint_config()
        self.endpoint_bookmark_field = "lastAccountAppraisalDate"



from concurrent import futures

from singer import get_logger, metrics
from singer.utils import strptime_to_utc

from .processor import TapProcessor
from ..helpers import convert
from ..helpers.exceptions import NoDeduplicationCapabilityException, NoDeduplicationKeyException

LOGGER = get_logger()


class DeduplicationProcessor(TapProcessor):
    def _init_endpoint_config(self):
        try:
            super(DeduplicationProcessor, self)._init_endpoint_config()
        except NoDeduplicationCapabilityException:
            pass
        self.endpoint_deduplication_key = "encoded_key"

        if len(self.generators) > 1:
            for generator in self.generators:
                if generator.endpoint_sorting_criteria.get("field") != self.endpoint_deduplication_key:
                    LOGGER.warning(f"(processor) Stream {self.stream_name} - Deduplication key should "
                                   "be the same as the sorting key in order for deduplication to work!")

    def _choose_next_record(self, generator_values):
        if not self.endpoint_deduplication_key:
            raise NoDeduplicationKeyException("Cannot continue deduplicating because the deduplication key is not set!")
        # Find lowest value in the list, and if two or more are equal, find max bookmark
        min_record_key = None
        min_record_value = None
        min_record_bookmark = None
        for iterator in generator_values:
            bookmark_field = convert(iterator.endpoint_bookmark_field)
            # Different record
            if min_record_key is None \
                    or min_record_value > generator_values[iterator][self.endpoint_deduplication_key]:
                min_record_key = iterator
                min_record_value = generator_values[iterator][self.endpoint_deduplication_key]
                if not bookmark_field or bookmark_field not in generator_values[iterator]:
                    continue
                min_record_bookmark = generator_values[iterator][bookmark_field]
            # Same record
            elif min_record_value == generator_values[iterator][self.endpoint_deduplication_key]:
                if not bookmark_field or bookmark_field not in generator_values[iterator]:
                    continue

                # Check the new bookmark against the min_record_key's bookmark
                min_record_dttm = strptime_to_utc(min_record_bookmark)
                record_dttm = strptime_to_utc(generator_values[iterator][bookmark_field])
                # If min_record has bookmark smaller than record, then replace min_record (we want highest bookmark)
                if min_record_dttm < record_dttm:
                    min_record_key = iterator
                    min_record_value = generator_values[iterator][self.endpoint_deduplication_key]
                    min_record_bookmark = generator_values[iterator][bookmark_field]

        return min_record_key, min_record_value

    def populate_generator_values(self):
        # Populate list of values from generators (if any were removed)
        for iterator in list(self.generator_values):
            if self.generator_values[iterator] is None:
                self.generator_values[iterator] = next(iterator, None)
            if self.generator_values[iterator] is None:
                self.generator_values.pop(iterator)

    def remove_same_key_values_from_generators(self, value):
        # Remove any record with the same deduplication_key from the list
        # (so we don't process the same record twice)
        for iterator in self.generator_values.keys():
            if value == self.generator_values[iterator][self.endpoint_deduplication_key]:
                self.generator_values[iterator] = None

    def process_records(self):
        record_count = 0
        with metrics.record_counter(self.stream_name) as counter:
            while True:
                self.populate_generator_values()
                if not self.generator_values:
                    break

                record_key, record_value = self._choose_next_record(self.generator_values)

                # Process the record
                record = self.generator_values[record_key]

                is_processed = self.process_record(record, record_key.time_extracted,
                                                    record_key.endpoint_bookmark_field)

                if is_processed:
                    record_count += 1
                    self._process_child_records(record)
                    counter.increment()

                self.remove_same_key_values_from_generators(record_value)
        return record_count



from . import setup_generator_base_test


def test_communications_generator_endpoint_config_init():
    generators = setup_generator_base_test("communications")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'communications/messages:search'
    assert generator.endpoint_params == {
        "detailsLevel": "FULL",
        "paginationDetails": "OFF"
    }
    assert generator.endpoint_bookmark_field == "creationDate"
    assert not hasattr(generator.endpoint_body, 'filterCriteria')
    assert not hasattr(generator.endpoint_body, 'sortingCriteria')



from datetime import datetime

import mock
from mock import MagicMock
from copy import deepcopy

from pytz import timezone

from . import setup_generator_base_test
from ..constants import config_json


@mock.patch("tap_mambu.tap_generators.audit_trail_generator.utc_now")
def test_audit_trail_generator_endpoint_config_init(utc_now_mock):
    fake_lte_date = datetime(year=2022, month=1, day=1,
                             hour=0, minute=0, second=0,
                             microsecond=0, tzinfo=timezone("UTC"))
    fake_lte_date_str = "2022-01-01T00:00:00.000000Z"
    utc_now_mock.return_value = fake_lte_date

    generators = setup_generator_base_test("audit_trail")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'v1/events'
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_api_version == "v1"
    assert generator.endpoint_api_key_type == "audit"
    assert generator.endpoint_params == {
        "sort_order": "asc",
        "occurred_at[gte]": "2021-06-01T00:00:00.000000Z",
        "occurred_at[lte]": fake_lte_date_str
    }
    assert generator.endpoint_bookmark_field == "occurred_at"
    assert generator.audit_trail_offset == 0


@mock.patch("tap_mambu.tap_generators.audit_trail_generator.get_bookmark")
@mock.patch("tap_mambu.tap_generators.audit_trail_generator.utc_now")
def test_audit_trail_generator_bookmark(utc_now_mock, get_bookmark_mock):
    fake_gte_date_str = "2021-09-01T00:00:00.000000Z"
    get_bookmark_mock.return_value = [fake_gte_date_str, 2]
    fake_lte_date = datetime(year=2022, month=1, day=1,
                             hour=0, minute=0, second=0,
                             microsecond=0, tzinfo=timezone("UTC"))
    fake_lte_date_str = "2022-01-01T00:00:00.000000Z"
    utc_now_mock.return_value = fake_lte_date

    generators = setup_generator_base_test("audit_trail")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.audit_trail_offset == 2
    assert generator.endpoint_params == {
        "sort_order": "asc",
        "occurred_at[gte]": fake_gte_date_str,
        "occurred_at[lte]": fake_lte_date_str
    }


def test_generator_iterator_flow():
    client_mock = MagicMock()
    client_mock.page_size = 5
    client_data = [[
        {"id": nr + page*5, "occurred_at": f"2021-10-01T00:00:{nr+page*5:02d}.000000Z"}
        for nr in range(5)] for page in range(3)]
    from tap_mambu.tap_generators.audit_trail_generator import AuditTrailGenerator
    generator = AuditTrailGenerator(stream_name="audit_trail",
                                    client=client_mock,
                                    config=config_json,
                                    state={"currently_syncing": "audit_trail"},
                                    sub_type="self")
    generator.prepare_batch = MagicMock()
    generator.fetch_batch = MagicMock(side_effect=client_data)

    expected_records = [
        {"custom_fields": [], "id": record["id"], "occurred_at": record["occurred_at"]}
        for record in deepcopy(sum(client_data, []))]
    record_count = 0
    for record in generator:
        assert record == expected_records[record_count]
        record_count += 1



from . import setup_generator_base_test


def test_groups_generator():
    generators = setup_generator_base_test("gl_accounts")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "glaccounts"
    assert generator.endpoint_bookmark_field == "lastModifiedDate"
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_params["type"] == generator.sub_type



from . import setup_generator_base_test


def test_loan_transactions_generator_endpoint_config_init():
    generators = setup_generator_base_test("loan_transactions")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'loans/transactions:search'
    assert generator.endpoint_sorting_criteria == {
        "field": "id",
        "order": "ASC"
    }
    assert generator.endpoint_bookmark_field == "creationDate"
    assert generator.endpoint_filter_criteria == [
        {
            "field": "creationDate",
            "operator": "AFTER",
            "value": '2021-06-01T00:00:00.000000Z'
        }
    ]



from .child_generator import ChildGenerator
from .no_pagination_generator import NoPaginationGenerator


class DepositCardsGenerator(NoPaginationGenerator, ChildGenerator):
    def _init_endpoint_config(self):
        super(DepositCardsGenerator, self)._init_endpoint_config()
        self.endpoint_api_method = "GET"
        self.endpoint_path = f"deposits/{self.endpoint_parent_id}/cards"
        self.endpoint_sorting_criteria = {}
        self.endpoint_filter_criteria = []



from .child_generator import ChildGenerator


class LoanRepaymentsGenerator(ChildGenerator):

    def _init_endpoint_config(self):
        super(LoanRepaymentsGenerator, self)._init_endpoint_config()
        self.endpoint_api_version = "v1"
        self.endpoint_api_method = "GET"
        # include parent id in endpoint path
        self.endpoint_path = f"loans/{self.endpoint_parent_id}/repayments"



from .multithreaded_bookmark_generator import MultithreadedBookmarkGenerator
from ..helpers.datetime_utils import datetime_to_utc_str


class DepositAccountsGenerator(MultithreadedBookmarkGenerator):
    def _init_config(self):
        super()._init_config()
        self.max_threads = 25

    def _init_endpoint_config(self):
        super(DepositAccountsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "deposits:search"
        self.endpoint_bookmark_field = "lastModifiedDate"
        self.endpoint_sorting_criteria = {
            "field": "lastModifiedDate",
            "order": "ASC"
        }

    def prepare_batch_params(self):
        super(DepositAccountsGenerator, self).prepare_batch_params()
        self.endpoint_filter_criteria[0]["value"] = datetime_to_utc_str(self.endpoint_intermediary_bookmark_value)



from pytz import timezone
from datetime import datetime
from tap_mambu import MambuClient
import dateutil.parser


_timezone = None
_datetime_formats = [
    "%Y-%m-%dT%H:%M:%S.%fZ%z",  # 0000-00-00T00:00:00.000000Z+00:00
    "%Y-%m-%dT%H:%M:%S.%f%z",   # 0000-00-00T00:00:00.000000+00:00
    "%Y-%m-%dT%H:%M:%SZ%z",     # 0000-00-00T00:00:00Z+00:00
    "%Y-%m-%dT%H:%M:%S%z",      # 0000-00-00T00:00:00+00:00
    "%Y-%m-%dT%H:%M%z",         # 0000-00-00T00:00+00:00

    "%Y-%m-%dT%H:%M:%S.%fZ",    # 0000-00-00T00:00:00.000000Z
    "%Y-%m-%dT%H:%M:%S.%f",     # 0000-00-00T00:00:00.000000
    "%Y-%m-%dT%H:%M:%SZ",       # 0000-00-00T00:00:00Z
    "%Y-%m-%dT%H:%M:%S",        # 0000-00-00T00:00:00
    "%Y-%m-%dT%H:%M",           # 0000-00-00T00:00

    "%Y-%m-%d",                 # 0000-00-00
]


def get_timezone_info(client: MambuClient):
    global _timezone
    response = client.request(method="GET", path="settings/organization", version="v1")
    _timezone = timezone(response.get("timeZoneID"))


def localize(dttm: datetime) -> datetime:
    if _timezone is None:
        raise RuntimeError("Cannot use timezone information without first calling 'get_timezone_info'")
    if dttm.tzinfo is None:  # If no timezone information is provided, we assume the datetime is in UTC format
        dttm = timezone("UTC").localize(dttm)
    # Convert datetime to Tenant Timezone
    return dttm.astimezone(_timezone)


def str_to_datetime(dttm_str: str) -> datetime:
    for datetime_format in _datetime_formats:
        try:
            return datetime.strptime(dttm_str, datetime_format)
        except ValueError:
            pass
    raise ValueError(f"Failed parsing datetime from string ({dttm_str})")


def str_to_localized_datetime(dttm_str: str) -> datetime:
    return localize(str_to_datetime(dttm_str))


def datetime_to_utc_str(dttm: datetime) -> str:
    return dttm.astimezone(timezone("UTC")).strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def datetime_to_local_str(dttm: datetime) -> str:
    return dttm.astimezone(_timezone).strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def datetime_to_tz(dttm: datetime, tz: str) -> datetime:
    return dttm.astimezone(timezone(tz))


def utc_now():
    return datetime.now(timezone("UTC"))


def local_now():
    return datetime.now(_timezone)



import datetime
from urllib.parse import parse_qs

import pytz

from mock import MagicMock, Mock

from tap_mambu.tap_generators.multithreaded_bookmark_generator import MultithreadedOffsetGenerator, \
    MultithreadedBookmarkGenerator, MultithreadedBookmarkDayByDayGenerator


class IsInstanceMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, self.expected_type)


class IsInstanceMockMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, MagicMock) and other.type == str(self.expected_type)


class GeneratorMock:
    def __init__(self, data):
        self.data = data
        self.time_extracted = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
        self.endpoint_bookmark_field = "last_modified_date"
        self.endpoint_sorting_criteria = {'field': 'id'}
        self.client = "mock_client"
        self.state = "mock_state"
        self.config = "mock_config"

    def __iter__(self):
        return self

    def __next__(self):
        if not self.data:
            raise StopIteration
        return self.data.pop(0)

    def write_bookmark(self):
        pass


class ClientMock:
    def __init__(self, page_size=100):
        self.page_size = page_size
        self.request = MagicMock()


class ClientWithDataMock(ClientMock):
    def __init__(self, page_size=100, bookmark_field="creationDate",
                 limit_field="limit", offset_field="offset", custom_data=None):
        super(ClientWithDataMock, self).__init__(page_size=page_size)
        self.base_url = "http://unit.test/api"
        self.request = MagicMock()
        self.limit_field = limit_field
        self.offset_field = offset_field
        self.data_to_serve = custom_data
        if self.data_to_serve is None:
            self.data_to_serve = [{"id": index, bookmark_field: f"2022-06-05T00:00:00.{index:06d}Z-07:00"}
                                  for index in range(400)] + \
                                 [{"id": index, bookmark_field: f"2022-06-06T00:00:00.{index:06d}Z-07:00"}
                                  for index in range(400)]
        self.request.side_effect = self.serve_request

    def serve_request(self, *args, **kwargs):
        params = kwargs.get("params")
        split_params = parse_qs(params)
        limit = int(split_params.get(self.limit_field, [None])[0])
        offset = int(split_params.get(self.offset_field, [None])[0])
        return self.data_to_serve[offset:limit+offset]


class MultithreadedOffsetGeneratorFake(MultithreadedOffsetGenerator):
    def __init__(self, stream_name='test_stream', client=ClientMock(), config=Mock(), state='', sub_type=''):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        super().__init__(stream_name, client, config, state, sub_type)


class MultithreadedBookmarkGeneratorFake(MultithreadedBookmarkGenerator):
    def __init__(self, stream_name='test_stream', client=ClientMock(), config=Mock(), state='', sub_type=''):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        super().__init__(stream_name, client, config, state, sub_type)


class MultithreadedBookmarkDayByDayGeneratorFake(MultithreadedBookmarkDayByDayGenerator):
    def __init__(self, stream_name='test_stream', client=ClientMock(), config=Mock(), state='', sub_type=''):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        super().__init__(stream_name, client, config, state, sub_type)



import mock
from mock import MagicMock
from . import setup_processor_base_test
from ..helpers import GeneratorMock
from ..constants import config_json


def test_audit_trail_processor():
    processor = setup_processor_base_test("audit_trail")

    assert processor.bookmark_offset == 1


@mock.patch("tap_mambu.helpers.write_state")
def test_audit_trail_update_bookmark(mock_write_state):
    from tap_mambu import discover
    from tap_mambu.tap_processors.audit_trail_processor import AuditTrailProcessor

    catalog = discover()
    client_mock = MagicMock()
    generator_mock = GeneratorMock([])
    generator_mock.static_params = {}
    processor = AuditTrailProcessor(catalog=catalog,
                                    stream_name="audit_trail",
                                    client=client_mock,
                                    config=config_json,
                                    state={'currently_syncing': 'audit_trail'},
                                    sub_type="self",
                                    generators=[generator_mock])

    assert processor.bookmark_offset == 1
    processor._update_bookmark({"occurred_at": "2021-10-01T00:00:00.000000Z"}, "occurred_at")
    assert processor.bookmark_offset == 1
    processor._update_bookmark({"occurred_at": "2021-10-01T00:00:00.000000Z"}, "occurred_at")
    assert processor.bookmark_offset == 2
    processor._update_bookmark({"occurred_at": "2021-10-01T00:00:00.000000Z"}, "occurred_at")
    assert processor.bookmark_offset == 3
    processor.write_bookmark()

    expected_state = {'currently_syncing': 'audit_trail', 'bookmarks': {
        'audit_trail': ['2021-10-01T00:00:00.000000Z', 3]}}
    mock_write_state.assert_called_once_with(expected_state)


@mock.patch("tap_mambu.helpers.write_state")
def test_audit_trail_write_bookmark(mock_write_state):
    from tap_mambu import discover
    from tap_mambu.tap_processors.audit_trail_processor import AuditTrailProcessor

    catalog = discover()
    client_mock = MagicMock()
    processor = AuditTrailProcessor(catalog=catalog,
                                    stream_name="audit_trail",
                                    client=client_mock,
                                    config=config_json,
                                    state={'currently_syncing': 'audit_trail'},
                                    sub_type="self",
                                    generators=[GeneratorMock([])])

    processor.write_bookmark()

    expected_state = {'currently_syncing': 'audit_trail', 'bookmarks': {
        'audit_trail': ['2021-06-01T00:00:00Z', 1]}}
    mock_write_state.assert_called_once_with(expected_state)



from .multithreaded_bookmark_generator import MultithreadedBookmarkGenerator
from ..helpers.datetime_utils import datetime_to_local_str, datetime_to_utc_str
from datetime import datetime


class CommunicationsGenerator(MultithreadedBookmarkGenerator):
    def _init_endpoint_config(self):
        super(CommunicationsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "communications/messages:search"
        self.endpoint_params = {
            "detailsLevel": "FULL",
            "paginationDetails": "OFF"
        }
        self.endpoint_bookmark_field = "creationDate"

    def modify_request_params(self, start, end):
        self.endpoint_body = [
            {
                "field": self.endpoint_bookmark_field,
                "operator": "AFTER",
                "value": datetime_to_utc_str(start)
            },
            {
                "field": self.endpoint_bookmark_field,
                "operator": "BEFORE",
                "value": datetime_to_utc_str(start)
            },
            {
                "field": "state",
                "operator": "EQUALS",
                "value": "SENT"
            }
        ]

    def _init_endpoint_body(self):
        self.endpoint_body = self.endpoint_filter_criteria

    def prepare_batch_params(self):
        super(CommunicationsGenerator, self).prepare_batch_params()
        self.endpoint_filter_criteria[0]["value"] = datetime_to_local_str(self.endpoint_intermediary_bookmark_value)



from mambu_tests.helpers import MultithreadedBookmarkDayByDayGeneratorFake
from tap_mambu.helpers.datetime_utils import datetime_to_tz, str_to_localized_datetime


def test_set_intermediary_bookmark_datetime():
    mock_datetime_record = datetime_to_tz(str_to_localized_datetime('2021-01-01T00:00:00.000000Z'), "UTC")

    generator = MultithreadedBookmarkDayByDayGeneratorFake()

    # test if the initial values aren't modified
    assert generator.endpoint_intermediary_bookmark_value is None
    assert generator.endpoint_intermediary_bookmark_offset == 0

    # check if the endpoint intermediary bookmark value and offset are set for a valid record with date time field
    generator.set_intermediary_bookmark(mock_datetime_record)
    assert generator.endpoint_intermediary_bookmark_value == mock_datetime_record.replace(hour=0, minute=0,
                                                                                          second=0, microsecond=0)
    assert generator.endpoint_intermediary_bookmark_offset == 1

    # check if the endpoint intermediary bookmark value isn't set with a value lower than the already saved one
    generator.set_intermediary_bookmark(datetime_to_tz(str_to_localized_datetime('2005-05-14T00:00:00.000000Z'), "UTC"))
    assert generator.endpoint_intermediary_bookmark_value == mock_datetime_record.replace(hour=0, minute=0,
                                                                                          second=0, microsecond=0)
    assert generator.endpoint_intermediary_bookmark_offset == 1

    # check if the endpoint intermediary bookmark value and offset correctly set when multiple records have
    # the same bookmark field value
    for _ in range(3):
        generator.set_intermediary_bookmark(mock_datetime_record)
    assert generator.endpoint_intermediary_bookmark_value == mock_datetime_record.replace(hour=0, minute=0,
                                                                                          second=0, microsecond=0)
    assert generator.endpoint_intermediary_bookmark_offset == 4

    # check if the endpoint intermediary bookmark value is properly set and the offset is reset
    mock_datetime_record = datetime_to_tz(str_to_localized_datetime('2021-11-11T00:00:00.000000Z'), "UTC")
    generator.set_intermediary_bookmark(mock_datetime_record)
    assert generator.endpoint_intermediary_bookmark_value == mock_datetime_record.replace(hour=0, minute=0,
                                                                                          second=0, microsecond=0)
    assert generator.endpoint_intermediary_bookmark_offset == 1


def test_set_intermediary_bookmark_date():
    mock_date_record = datetime_to_tz(str_to_localized_datetime('2020-01-01'), "UTC")

    generator = MultithreadedBookmarkDayByDayGeneratorFake()

    # test if the initial values aren't modified
    assert generator.endpoint_intermediary_bookmark_value is None
    assert generator.endpoint_intermediary_bookmark_offset == 0

    # check if the endpoint intermediary bookmark value and offset are set for a valid record with date field
    generator.set_intermediary_bookmark(mock_date_record)
    assert generator.endpoint_intermediary_bookmark_value == mock_date_record
    assert generator.endpoint_intermediary_bookmark_offset == 1

    # check if the endpoint intermediary bookmark value isn't set with a value lower than the already saved one
    generator.set_intermediary_bookmark(datetime_to_tz(str_to_localized_datetime('2005-05-14'), "UTC"))
    assert generator.endpoint_intermediary_bookmark_value == mock_date_record
    assert generator.endpoint_intermediary_bookmark_offset == 1

    # check if the endpoint intermediary bookmark value and offset correctly set when multiple records have
    # the same bookmark field value
    for _ in range(3):
        generator.set_intermediary_bookmark(mock_date_record)
    assert generator.endpoint_intermediary_bookmark_value == mock_date_record
    assert generator.endpoint_intermediary_bookmark_offset == 4

    # check if the endpoint intermediary bookmark value is properly set and the offset is reset
    mock_date_record = datetime_to_tz(str_to_localized_datetime('2021-11-11'), "UTC")
    generator.set_intermediary_bookmark(mock_date_record)
    assert generator.endpoint_intermediary_bookmark_value == mock_date_record
    assert generator.endpoint_intermediary_bookmark_offset == 1



import singer
from concurrent.futures import Future, ThreadPoolExecutor
from typing import List


LOGGER = singer.get_logger()


class MultithreadedRequestsPool:
    _dispatcher = ThreadPoolExecutor(max_workers=20)

    @classmethod
    def shutdown(cls):
        cls._dispatcher.shutdown(wait=True, cancel_futures=True)

    @staticmethod
    def run(client, stream_name,
            endpoint_path, endpoint_api_method, endpoint_api_version,
            endpoint_api_key_type, endpoint_body, endpoint_params) -> List[dict]:
        endpoint_querystring = '&'.join([f'{key}={value}' for (key, value) in endpoint_params.items()])

        LOGGER.info(f'(generator) Stream {stream_name} - URL for {stream_name} ({endpoint_api_method}, '
                    f'{endpoint_api_version}): {client.base_url}/{endpoint_path}?{endpoint_querystring}'
                    f' - body = {endpoint_body}')

        response = client.request(
            method=endpoint_api_method,
            path=endpoint_path,
            version=endpoint_api_version,
            apikey_type=endpoint_api_key_type,
            params=endpoint_querystring,
            endpoint=stream_name,
            json=endpoint_body
        )

        LOGGER.info(f'(generator) Stream {stream_name} - extracted records: {len(response)}')
        return response

    @classmethod
    def queue_request(cls, client, stream_name,
                      endpoint_path, endpoint_api_method, endpoint_api_version,
                      endpoint_api_key_type, endpoint_body, endpoint_params) -> Future:
        return cls._dispatcher.submit(cls.run, client, stream_name, endpoint_path, endpoint_api_method,
                                      endpoint_api_version, endpoint_api_key_type, endpoint_body, endpoint_params)

    @classmethod
    def queue_function(cls, func, *args, **kwargs):
        return cls._dispatcher.submit(func, *args, **kwargs)



from .multithreaded_offset_generator import MultithreadedOffsetGenerator


class UsersGenerator(MultithreadedOffsetGenerator):
    def __init__(self, stream_name, client, config, state, sub_type):
        super(UsersGenerator, self).__init__(stream_name, client, config, state, sub_type)
        self.date_windowing = False

    def _init_endpoint_config(self):
        super(UsersGenerator, self)._init_endpoint_config()
        self.endpoint_path = "users"
        self.endpoint_api_method = "GET"
        self.endpoint_params["sortBy"] = "id:ASC"
        self.endpoint_bookmark_field = "lastModifiedDate"



from .child_processor import ChildProcessor


class LoanRepaymentsProcessor(ChildProcessor):
    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_parent = "loan_accounts"



_threads_original_shutdown = None


def pytest_collection(session):
    import sys
    from mock import MagicMock

    def mock_on_exception(*args, **kwargs):
        def retry(func):
            return func

        return retry

    # this assignment will replace the backoff lib for all tests from this file, but this is acceptable because
    # there isn't any need of its functionality
    sys.modules['backoff'] = MagicMock()
    sys.modules['backoff'].on_exception = mock_on_exception

    from tap_mambu.helpers.multithreaded_requests import MultithreadedRequestsPool

    global _threads_original_shutdown
    _threads_original_shutdown = MultithreadedRequestsPool.shutdown
    MultithreadedRequestsPool.shutdown = MagicMock()


    from pytz import timezone
    from tap_mambu.helpers import datetime_utils
    datetime_utils._timezone = timezone("US/Pacific")


def pytest_sessionfinish(session, exitstatus):
    _threads_original_shutdown()



from .multithreaded_bookmark_generator import MultithreadedBookmarkDayByDayGenerator
from ..helpers.datetime_utils import datetime_to_utc_str
from datetime import datetime


class ActivitiesGenerator(MultithreadedBookmarkDayByDayGenerator):
    def _init_endpoint_config(self):
        super(ActivitiesGenerator, self)._init_endpoint_config()
        self.endpoint_path = "activities"
        self.endpoint_api_method = "GET"
        self.endpoint_api_version = "v1"
        self.endpoint_bookmark_field = "timestamp"

    def modify_request_params(self, start, end):
        self.static_params["from"] = datetime.strftime(start, '%Y-%m-%d')
        self.static_params["to"] = datetime.strftime(end, '%Y-%m-%d')

    @staticmethod
    def unpack_activity(record):
        record.update(record["activity"])
        del record["activity"]
        return record

    def transform_batch(self, batch):
        return list(map(self.unpack_activity, super(ActivitiesGenerator, self).transform_batch(batch)))

    def prepare_batch_params(self):
        super(ActivitiesGenerator, self).prepare_batch_params()
        self.static_params['to'] = datetime_to_utc_str(self.endpoint_intermediary_bookmark_value)[:10]

    def compare_bookmark_values(self, a, b):
        return a < b



from .generator import TapGenerator
from typing import List


class ChildGenerator(TapGenerator):
    def __init__(self, stream_name, client, config, state, sub_type, parent_id):
        self.endpoint_parent_id = parent_id
        super(ChildGenerator, self).__init__(stream_name, client, config, state, sub_type)

    def _init_endpoint_config(self):
        super(ChildGenerator, self)._init_endpoint_config()
        self.endpoint_path = f"{self.endpoint_parent_id}"  # include parent id in endpoint path

    def _init_buffers(self):
        self.buffer: List = list()
        self.max_buffer_size = 1000

